# -*- coding: utf-8 -*-
##############################################################################
#
#    wms module for OpenERP, WMS Extended for Open ERP
#    Copyright (C) 2010 SYLEAM Info Services (<http://www.syleam.fr/>)
#              Christophe CHAUVET <christophe.chauvet@syleam.fr>
#
#    This file is a part of wms
#
#    wms is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    wms is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv
from osv import fields
from tools.translate import _


class StockPicking(osv.osv):
    """
    Add crossdock process
    """
    _inherit = 'stock.picking'

    def _action_type(self, cr, uid, context=None):
        """
        Inherit properly the list of type
        """
        if context is None:
            context = {}
        res = super(StockPicking, self)._columns['type'].selection
        if 'transfer' not in [k for k, v in res]:
            res.append(('transfer', 'Transfer'))
        return res

    _columns = {
        'type': fields.selection(_action_type, 'Type', required=True),
        'warehouse_id': fields.many2one('stock.warehouse', 'Warehouse', help='Indicate the warehouse for this picking'),
        'picking_time': fields.float('Time', help='Time to made this picking'),
        'transporter_id': fields.many2one('res.partner', 'Transporter'),
        'transporter_number': fields.char('Tracking number', size=128),
        'product_id': fields.related('move_lines', 'product_id', type='many2one', relation='product.product', string='Product'),
        'round_id': fields.many2one('stock.round', 'Round', help='Use for manage deliveries'),
    }

    _defaults = {
        'warehouse_id': lambda obj, cr, uid, context: obj.pool.get('res.users')._default_warehouse(cr, uid, context),
        'picking_time': lambda *a: 0.0,
        'transporter_id': lambda *a: False,
        'transporter_number': lambda *a: False,
        'round_id': lambda *a: False,
    }

    def default_get(self, cr, uid, fields_list, context=None):
        if context is None:
            context = {}
        values = super(StockPicking, self).default_get(cr, uid, fields_list, context=context)
        if 'type' in fields_list and values.get('type') == 'transfer' and values.get('name', '/') == '/':
            values['name'] = self.pool.get('ir.sequence').get(cr, uid, 'stock.picking.transfer')

        return values

    def onchange_wms(self, cr, uid, ids, address_id, context=None):
        """
        This on_change manage all values for WMS module
        """
        if context is None:
            context = {}

        vals = {}
        if address_id:
            addr = self.pool.get('res.partner.address').browse(cr, uid, address_id, context=context)
            if addr.round_id:
                vals['round_id'] = addr.round_id.id
            elif addr.partner_id and addr.partner_id.round_id:
                vals['round_id'] = addr.partner_id.round_id.id
        else:
            vals['round_id'] = False

        return {'value': vals}

    def create(self, cr, uid, values, context=None):
        """
        When picking is create, check if there is an address, and retrieve round value
        if round is on address, take it or on partner
        """
        if context is None:
            context = {}

        location_id = False
        if values.get('address_id') and not values.get('round_id', False) and not context.get('disable_round', False):
            addr = self.pool.get('res.partner.address').browse(cr, uid, values.get('address_id'), context=context)
            if addr.round_id:
                location_id = addr.round_id.default_location_id.id
                values['round_id'] = addr.round_id.id
            elif addr.partner_id and addr.partner_id.round_id:
                location_id = addr.partner_id.round_id.default_location_id.id
                values['round_id'] = addr.partner_id.round_id.id

        return super(StockPicking, self).create(cr, uid, values, context=context)

    def action_confirm(self, cr, uid, ids, context=None):
        """
        Change destination location on move, if there is a round on the picking
        """
        move_obj = self.pool.get('stock.move')
        pick = self.browse(cr, uid, ids, context=context)
        for p in pick:
            if p.type == 'out' and p.state == 'draft' and p.round_id and p.round_id.default_location_id:
                move_ids = [x.id for x in p.move_lines]
                move_obj.write(cr, uid, move_ids, {'location_dest_id': p.round_id.default_location_id.id}, context=context)

        return super(StockPicking, self).action_confirm(cr, uid, ids, context=context)

    def product_crossdock_location(self, cr, uid, product_id, warehouse_id, context=None):
        """
        Search id 
        """
        if context is None:
            context = {}

        loc_ids = self.pool.get('stock.location').search(cr, uid, [('warehouse_id', '=', warehouse_id), ('usage', '=', 'crossdock')], context=context)
        move_ids = self.pool.get('wms.report.stock.available').search(cr, uid, [('product_id', '=', product_id), ('location_id', 'in', loc_ids)], context=context)
        return [x.location_id.id for x in self.pool.get('wms.report.stock.available').browse(cr, uid, move_ids, context=context)]

    def crossdock_location_available(self, cr, uid, warehouse_ids, context=None):
        """
        Return the list ids of empty crossdock location for these warehouse
        You can specify only_one in context to retrieve 1 location at the time

        :param warehouse_ids: list of the warehouse
        :type  warehouse_ids: list
        :return: dict with the list of empty crossdock location
        :rtype: dict
        """
        if context is None:
            context = {}

        st_av_obj = self.pool.get('wms.report.stock.available')
        res = {}
        for w in warehouse_ids:
            tmp_ids = []
            l_ids = self.pool.get('stock.location').search(cr, uid,
                        [('warehouse_id', '=', w), ('usage', '=', 'crossdock')], context=context)
            for a in l_ids:
                if not st_av_obj.search(cr, uid, [('location_id', '=', a)]):
                    tmp_ids.append(a)
                if tmp_ids and context.get('only_one'):
                    break

            if not tmp_ids:
                default_crossdock = self.pool.get('stock.warehouse').read(cr, uid, w, ['crossdock_location_id'], context=context)['crossdock_location_id']
                if default_crossdock:
                    tmp_ids = [default_crossdock[0]]
            res[str(w)] = tmp_ids
        return res

    def create_crossdock_move(self, cr, uid, move, crossdock_id=None, context=None):
        """
        Check if the product on this stock.move have been reserved,
        and affecte the location on the picking_order

        :param move_id: movement to check
        :type  move_id: int
        :return: True if the movement is correct
        :rtype: boolean
        """
        if context is None:
            context = {}

        move_obj = self.pool.get('stock.move')

        if crossdock_id is None:
            default_crossdock = self.pool.get('stock.warehouse').read(cr, uid, move.location_dest_id.warehouse_id.id,
                                        ['crossdock_location_id'], context=context)['crossdock_location_id']
            if not default_crossdock:
                raise osv.except_osv(_('Error'), _('No default crossdock location defined on the warehouse!'))
            crossdock_id = isinstance(default_crossdock, list) and default_crossdock[0] or default_crossdock

        ###
        ## If we have an outgoing quantity, and location type is crossdock, we split in two line
        ##
        new_id = False
        if abs(move.product_id.outgoing_qty) > 0:

            # Search how many quantity must be put in crossdock
            picking_ids = self.search(cr, uid, [('type', '=', 'out'), ('state', 'in', ['confirmed', 'assigned'])],
                                context=context)
            args = [
                ('picking_id', 'in', picking_ids),
                ('product_id', '=', move.product_id.id),
                ('state', 'in', ['confirmed', 'waiting']),
            ]
            reserve_ids = move_obj.search(cr, uid, args,
                                limit=abs(move.product_qty), order='date_planned', context=context)

            reserves = move_obj.browse(cr, uid, reserve_ids, context=context)
            total_crossdock = 0.0
            for r in reserves:
                total_crossdock += r.product_qty

            # Check if we must split the line in two if there is enougth quantity available
            if move.product_qty > total_crossdock:
                args = {
                    'product_qty': (move.product_qty - total_crossdock),
                    'state': 'done',
                    'crossdock_ok': True,
                }
                new_id = move_obj.copy(cr, uid, move.id, args, context=context)
                args = {
                    'product_qty': total_crossdock,
                    'location_dest_id': crossdock_id,
                    'state': 'done',
                    'crossdock_ok': True,
                }
                move_obj.write(cr, uid, [move.id], args, context=context)
            else:
                # We don't have enougth quantity, how many moves can be assigned with quantity
                args = {
                    'location_dest_id': crossdock_id,
                    'state': 'done',
                    'crossdock_ok': True,
                }
                move_obj.write(cr, uid, [move.id], args, context=context)

                reserve_ids = []
                tmp_qty = 0.0
                for r in reserves:
                    tmp_qty += r.product_qty
                    if tmp_qty <= move.product_qty:
                        reserve_ids.append(r.id)
                    else:
                        break

            args = {
                'location_id': crossdock_id,
                'state': 'assigned',
            }
            move_obj.write(cr, uid, reserve_ids, args, context=context)

        return new_id or move.id

    def test_finnished(self, cr, uid, ids):
        """
        On receipt order, check if there is picking, and generate crossdock
        1) check if all move are finnished
        2) if not return False
        """
        in_ids = []
        other_ids = []
        for p in self.browse(cr, uid, ids):
            if p.type == 'in':
                in_ids.append(p.id)
            else:
                other_ids.append(p.id)

        # check if there is a receipt order, if not call directly the super method
        if len(in_ids) == 0:
            return super(StockPicking, self).test_finnished(cr, uid, ids)

        # if we have a receipt we must check if product have been reserved
        elif len(in_ids) >= 1:
            move_obj = self.pool.get('stock.move')
            lines_ok = True
            move_ids = move_obj.search(cr, uid, [('picking_id', 'in', in_ids)])
            cds_ids = []
            new_move_ids = []
            moves = move_obj.browse(cr, uid, move_ids)
            for move in moves:
                if not cds_ids and move.location_dest_id.warehouse_id.id:
                    cd_ids = self.crossdock_location_available(cr, uid, [move.location_dest_id.warehouse_id.id])
                    if isinstance(cd_ids, dict):
                        cd_ids = cd_ids[str(move.location_dest_id.warehouse_id.id)]

                if move.state not in ('done', 'cancel'):
                    if move.product_qty != 0.0:
                        lines_ok = False

            if not lines_ok:
                return False

            for move in moves:
                # check if product is in crossdock mode and if out picking is available
                if not move.crossdock_ok and move.product_id.location_type == 'crossdock' and move.product_qty > 0:

                    pro_crossdock = self.product_crossdock_location(cr, uid, move.product_id.id,
                                        move.location_dest_id.warehouse_id.id)
                    if pro_crossdock:
                        tmp_crossdock = pro_crossdock[0]
                    else:
                        tmp_crossdock = cd_ids and cd_ids.pop() or None
                    m_id = self.create_crossdock_move(cr, uid, move, crossdock_id=tmp_crossdock)
                    new_move_ids.append(m_id)

            #
            # For each picking out in confirm, check all lines and verify if it available
            # If yes, execute the workflow.
            args = [
                ('type','=','out'),
                ('state','=','confirmed'),
            ]
            p_ids = self.search(cr, uid, args)
            pick_list = []
            for p in self.browse(cr, uid, p_ids):
                l_ok = True
                for l in p.move_lines:
                    if l.state not in ('assigned','cancel'):
                        l_ok = False

                if l_ok:
                    pick_list.append(p.id)

            for i in pick_list:
                self.force_assign(cr, uid, [i])

        return super(StockPicking, self).test_finnished(cr, uid, ids)

    def move_merge(self, cr, uid, picking_id, context=None):
        """
        Merge move lines, keep lot, tracking
        Returns move lines in dico : {'product_key': (quantity, move_line_id)}
        """
        if context is None:
            context = {}

        move_obj = self.pool.get('stock.move')
        move_ids = move_obj.search(cr, uid, [('picking_id', '=', picking_id)], context=context)
        known_moves = {}
        moves_to_delete = []
        for move in move_obj.browse(cr, uid, move_ids, context=context):
            product_key = "%d_%d_%d" % (move.product_id.id, move.prodlot_id and move.prodlot_id.id or 0, move.tracking_id and move.tracking_id.id or 0)
            # Check if product_id is known
            if product_key in known_moves.keys():
                # Add quantities
                known_moves[product_key].product_qty += move.product_qty
                moves_to_delete.append(move.id)
            else:
                known_moves[product_key] = move

        # Saves modified moves
        for p, move in known_moves.items():
            move_obj.write(cr, uid, [move.id], {'product_qty': move.product_qty}, context=context)
            known_moves[p] = (move.product_qty, move.id, )

        # Delete moves to delete
        move_obj.unlink(cr, uid, moves_to_delete, context=context)

        return known_moves

StockPicking()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
