# -*- coding: utf-8 -*-
##############################################################################
#
#    wms_scanner module for OpenERP, Module for manage barcode reader
#    Copyright (C) 2010 SYLEAM (<http://syleam.fr/>)
#              Christophe Chauvet <christophe.chauvet@syleam.fr>
#
#    This file is a part of wms_scanner
#
#    wms_scanner is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    wms_scanner is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from osv import osv
from osv import fields
from tools.translate import _
import netsvc

logger = netsvc.Logger()


class ScannerScenario(osv.osv):
    _name = 'scanner.scenario'
    _description = 'Scenario for scanner'

    _columns = {
        'name': fields.char('Name', size=64, ),
        'sequence': fields.integer('Sequence'),
        'active': fields.boolean('Active', help='if check, this object is always available'),
        'model_id': fields.many2one('ir.model', 'Model', required=True),
        'step_ids': fields.one2many('scanner.scenario.step', 'scenario_id', 'Scenario', ),
        'warehouse_ids': fields.many2many('stock.warehouse', 'scanner_scenario_warehouse_rel', 'scenario_id', 'warehouse_id', 'Warehouses'),
    }

ScannerScenario()


class ScannerScenarioStep(osv.osv):
    _name = 'scanner.scenario.step'
    _description = 'Step for scenario'

    _columns = {
        'name': fields.char('Name', size=64, help='Name of the step'),
        'scenario_id': fields.many2one('scanner.scenario', 'Scenario', required=True),
        'step_start': fields.boolean('Step start',help='Check this if this is the first step of scenario'),
        'step_stop': fields.boolean('Step stop',help='check this if this is the last step of scenario'),
        'out_trans_ids': fields.one2many('scanner.scenario.transition', 'from_id', 'Outgoing transitions'),
        'in_trans_ids': fields.one2many('scanner.scenario.transition', 'to_id', 'Incoming transitions'),
        'python_code': fields.text('Python code', ),
    }

    _defaults = {
        'step_start': lambda *a: False,
        'step_stop': lambda *a: False,
        'python_code': lambda *a: '# Use message to retrieve the data transmit by the scanner\n# res containt the result',
    }

ScannerScenarioStep()


class ScannerScenarioTransition(osv.osv):
    _name = 'scanner.scenario.transition'
    _description = 'Transition for scenario'

    _columns = {
        'name': fields.char('Name', size=64, required=True),
        'from_id': fields.many2one('scanner.scenario.step', 'From', required=True),
        'to_id': fields.many2one('scanner.scenario.step', 'To', required=True),
        'condition': fields.char('Condition', size=256, required=True),
        'trans_type': fields.selection([('scan','Scan'),('key','Keyboard')], 'Transition Type', ),
        'tracer': fields.char('Tracer', size=12, help='Use to determine, on which transition we arrive to this step'),
    }

    _defaults = {
        'condition': lambda *a: 'True',
        'trans_type': lambda *a: 'scan',
        'tracer': lambda *a: False,
    }

ScannerScenarioTransition()


class ScannerHardware(osv.osv):
    _name = 'scanner.hardware'
    _description = 'Scanner Hardware'

    _columns = {
        'name': fields.char('Name', size=128, required=True),
        'code': fields.char('Code', size=12, required=True),
        'warehouse_id': fields.many2one('stock.warehouse', 'Warehouse', required=True),
        'scenario_id': fields.many2one('scanner.scenario', 'Scenario'),
        'step_id': fields.many2one('scanner.scenario.step', 'Current Step'),
        'ref_doc': fields.integer('Reference'),
        'tmp_val1': fields.char('Temp value 1', size=128, ),
        'tmp_val2': fields.char('Temp value 2', size=128, ),
        'tmp_val3': fields.char('Temp value 3', size=128, ),
        'tmp_val4': fields.char('Temp value 4', size=128, ),
        'tmp_val5': fields.char('Temp value 5', size=128, ),
    }

    _defaults = {
        'scenario_id': lambda *a: False,
        'step_id': lambda *a: False,
        'ref_doc': lambda *a: False,
    }

    def scanner_check(self, cr, uid, numterm):
        """
        Check the step for this scanner
        """
        context = self.pool.get('res.users').context_get(cr, uid)
        term_id = self.search(cr, uid, [('code','=',numterm)])
        if not term_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'Terminal %s not exists!' % numterm) 
            raise osv.except_osv(_('Error'), _('This terminal not exists!'))
        term = self.browse(cr, uid, term_id[0])
        if term.scenario_id:
            return True
        return False

    def scanner_call(self, cr, uid, numterm=None, action=None, message='', etype='K'):
        """
        This method is the only one, that called by the barcode reader,

        :param etype: Indicate K for keyboard, S for Scan, use for transition
        :type  etype: str
        :return: A tuple for the action
        :rtype: tuple
        """
        context = self.pool.get('res.users').context_get(cr, uid)
        logger.notifyChannel('wms_scanner', netsvc.LOG_DEBUG, 'SCANNER_CALL')
        if numterm is None:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'scanner_call required numterm argument') 
            raise osv.except_osv(_('Error'), _('Terminal number is required'))

        if action is None:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'scanner_call required action argument') 
            raise osv.except_osv(_('Error'), _('Action is required'))

        term_id = self.search(cr, uid, [('code','=',numterm)])
        if not len(term_id):
            raise osv.except_osv(_('Error'), _('This terminal is not declare!'))

        scenario = self.pool.get('scanner.scenario')
        term = self.browse(cr, uid, term_id[0])

        # we affect this terminal to the related action
        # search the schnario and the first step
        if action == 'action' and not term.scenario_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, '[%s] Action : %s (pas de scenario en cours)' % (numterm, message))
            r = self._scenario_save(cr, uid, term.id, message, etype)
            return r
        elif action == 'action' and term.scenario_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, '[%s] Action : %s (%s)' % (numterm, message, term.scenario_id.name))
            r = self._scenario_save(cr, uid, term.id, message, etype, term.scenario_id.id, term.step_id.id, term.ref_doc)
            return r

        # check id terminal is not affected to a scenario and step
        # If not send the menu
        if not term.scenario_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, '[%s] Pas de scenario' % numterm) 
            return self._scenario_list(cr, uid, term.warehouse_id.id)

        return ('M', ['Unknown','action'], '')


    def scanner_end(self, cr, uid, numterm=None, context=None):
        """
        End the end barcode is read, we execute this step
        """
        if context is None:
            context = self.pool.get('res.users').context_get(cr, uid)

        term_id = self.search(cr, uid, [('code','=',numterm)])
        if not term_id:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'Terminal %s not exists!' % numterm) 
            raise osv.except_osv(_('Error'), _('This terminal not exists!'))

        logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, 'End scenario request for %s terminal' % numterm) 
        self.write(cr, uid, term_id, {'scenario_id': False, 'step_id': False, 'ref_doc': False})

        return ('F', ['This scenario','is finished'], '')

    def _memorize(self, cr, uid, term_id, scenario_id, step_id, object=None, context=None):
        """
        After affect a scenario to a scanner, we must memorize it
        If object is specify, save it as well (ex: res.partner,12)
        """
        if context is None:
            context = self.pool.get('res.users').context_get(cr, uid)

        sca = self.pool.get('scanner.hardware')
        sca_args = {'scenario_id': scenario_id, 'step_id': step_id}
        if object and isinstance(object, int):
            sca_args['ref_doc'] = object
        sca.write(cr, uid, [term_id], sca_args)

    def _scenario_save(self, cr, uid, term_id, message, etype, scenario_id=None, step_id=None, cur_obj='', context=None):
        """
        Save the scenario on this terminal and execute the current step
        Return the action to the terminal
        """
        if context is None:
            context = self.pool.get('res.users').context_get(cr, uid)

        if etype == 'K':
            etype = 'key'
        else:
            etype = 'scan'

        scenario = self.pool.get('scanner.scenario')
        step = self.pool.get('scanner.scenario.step')
        term = self.pool.get('scanner.hardware')
        old_step = False
        # No scenario, ok we search it
        tracer = False
        if not scenario_id:
            scenario_ids = scenario.search(cr, uid, [('name','=', message)])
            if scenario_ids:
                scenario_id = scenario_ids[0]
                step_ids = step.search(cr, uid, [('scenario_id','=',scenario_id),('step_start','=',True)])
                if not step_ids:
                    return ('M', ['please contact','your','administrator','A001'], '')
                step_id = step_ids[0]
        else:
            # we search the next step for the current scenario
            old_step = step_id
            trans = self.pool.get('scanner.scenario.transition')
            trs_ids = trans.search(cr, uid, [('from_id', '=', step_id)])
            if not trs_ids:
                return ('M', ['Please contact','your','administrator'], '')
            tr_vals = trans.read(cr, uid, trs_ids[0], ['to_id','tracer'])
            step_id = tr_vals['to_id'][0]
            tracer = tr_vals['tracer']

        self._memorize(cr, uid, term_id, scenario_id, step_id, cur_obj, context)

        # execute the step and return the result
        stp = step.browse(cr, uid, step_id, context=context)
        logger.notifyChannel('wms_scanner', netsvc.LOG_INFO, 'Model: %s' % stp.scenario_id.model_id.model)
        ld = {
            'cr': cr, 
            'uid': uid, 
            'pool': self.pool,
            'model': self.pool.get(stp.scenario_id.model_id.model), 
            'term': term,
            'context': context,
            'm': message,
            't': term.browse(cr, uid, term_id, context=context),
            'tracer': tracer,
            'wkf': netsvc.LocalService('workflow'),
        }
        try:
            exec stp.python_code in ld
        except osv.except_osv, e:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'OSV Exception: %s' % str(e)) 
            return ('M', ['Please contact','your','administrator'], '')
        except Exception, e:
            logger.notifyChannel('wms_scanner', netsvc.LOG_WARNING, 'OSV Exception: %s' % str(e)) 
            return ('M', ['Please contact','your','administrator'], '')
        return (ld.get('act', 'M'), ld.get('res',['nothing']), ld.get('val', '0'))

    def _scenario_list(self, cr, uid, warehouse_id):
        """
        Retrieve the scenario list for this warehouse
        """
        # List scenario for this material
        context = self.pool.get('res.users').context_get(cr, uid)
        scenario = self.pool.get('scanner.scenario')
        scenario_ids = scenario.search(cr, uid, [('warehouse_ids','in', [warehouse_id])])
        scene = scenario.read(cr, uid, scenario_ids, ['name'], context=context)

        ret = []
        for s in scene:
            ret.append(s['name'])
        return ('M', ret, '')

ScannerHardware()



# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
